# -*- coding: utf-8 -*-
# file: __init__.py
# time: 14:53 06/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.

"""
This __init__.py file exposes the Key API Entries of the library for easy access.
Use dir(omnigenbench) to see all available APIs.

Key API Entries:
----------------
- AutoBench: Automated benchmarking of genomic models
- AutoTrain: Automated training of genomic models
- BenchHub: Hub for accessing benchmarks
- ModelHub: Hub for accessing pre-trained models
- PipelineHub: Hub for accessing pipelines
- Various model classes for different genomic tasks
- Dataset classes for different data formats
- Tokenizer classes for different sequence representations
- Metric classes for evaluation
- Trainer classes for model training
"""

__name__ = "omnigenbench"
__version__ = "0.3.10alpha"
__author__ = "YANG, HENG"
__email__ = "yangheng2021@gmail.com"
__license__ = "Apache-2.0"

# Import core auto components
from .auto.auto_bench.auto_bench import AutoBench
from .auto.config.auto_config import AutoConfig
from .auto.bench_hub.bench_hub import BenchHub
from .auto.auto_train.auto_train import AutoTrain
from .auto.auto_bench.auto_bench_cli import run_bench, bench_command
from .auto.auto_train.auto_train_cli import run_train, train_command

# Import source modules
from .src import dataset, metric, model, tokenizer

# Import abstract base classes
from .src.abc.abstract_dataset import OmniDataset
from .src.abc.abstract_metric import OmniMetric
from .src.abc.abstract_model import OmniModel
from .src.abc.abstract_tokenizer import OmniTokenizer

# Import dataset classes
from .src.dataset.omni_dataset import (
    OmniDatasetForSequenceClassification,
    OmniDatasetForSequenceRegression,
    OmniDatasetForTokenClassification,
    OmniDatasetForTokenRegression,
)


# Import metric classes
from .src.metric import ClassificationMetric, RegressionMetric, RankingMetric

# Import utility functions
from .src.misc.utils import (
    clean_temp_dir_pt_files,
    fprint,
    seed_everything,
    save_args,
    naive_secondary_structure_repair,
    check_bench_version,
    clean_temp_checkpoint,
    print_args,
    env_meta_info,
    RNA2StructureCache,
)

# Import model classes
from .src.model import (
    OmniModelForSequenceClassification,
    OmniModelForMultiLabelSequenceClassification,
    OmniModelForTokenClassification,
    OmniModelForSequenceRegression,
    OmniModelForTokenRegression,
    OmniModelForStructuralImputation,
    OmniModelForMatrixRegression,
    OmniModelForMatrixClassification,
    OmniModelForMLM,
    OmniModelForSeq2Seq,
    OmniModelForRNADesign,
    OmniModelForEmbedding,
    OmniModelForAugmentation,
)

# Import LoRA model
from .src.lora.lora_model import OmniLoraModel

# Import tokenizer classes
from .src.tokenizer import (
    OmniBPETokenizer,
    OmniKmersTokenizer,
    OmniSingleNucleotideTokenizer,
)

# Import trainer classes
from .src.trainer.hf_trainer import HFTrainer
from .src.trainer.trainer import Trainer
from .src.trainer.accelerate_trainer import AccelerateTrainer

# Import hub utilities
from .utility.hub_utils import (
    download_benchmark,
    download_model,
    download_pipeline,
    query_models_info,
)
from .utility import hub_utils

# Import hub classes
from .utility.model_hub.model_hub import ModelHub
from .utility.dataset_hub.dataset_hub import load_benchmark_datasets
from .utility.pipeline_hub.pipeline import Pipeline
from .utility.pipeline_hub.pipeline_hub import PipelineHub

# Import module utilities
from .src.model.module_utils import OmniPooling
from .utility.ensemble import VoteEnsemblePredictor

# --------------------------------------------------------------------------------
# For backward compatibility version 0.2.7alpha and earlier
from .auto.config.auto_config import AutoBenchConfig

# Import explainer classes
from .src.explainability.epistasis.explainer import EpistasisExplainer
from .src.explainability.sequence_logo.explainer import SequenceLogoExplainer
from .src.explainability.visualization_2d.explainer import Visualization2DExplainer

OmniGenomeTokenizer = OmniTokenizer
OmniGenomeKmersTokenizer = OmniKmersTokenizer
OmniGenomeSingleNucleotideTokenizer = OmniSingleNucleotideTokenizer
OmniGenomeBPETokenizer = OmniBPETokenizer
OmniGenomeDataset = OmniDataset
OmniGenomeMetric = OmniMetric
OmniGenomeModel = OmniModel
OmniGenomeDatasetForSequenceClassification = OmniDatasetForSequenceClassification
OmniGenomeDatasetForSequenceRegression = OmniDatasetForSequenceRegression
OmniGenomeDatasetForTokenClassification = OmniDatasetForTokenClassification
OmniGenomeDatasetForTokenRegression = OmniDatasetForTokenRegression
OmniGenomeLoraModel = OmniLoraModel
OmniGenomeModelForSequenceClassification = OmniModelForSequenceClassification
OmniGenomeModelForMultiLabelSequenceClassification = (
    OmniModelForMultiLabelSequenceClassification
)
OmniGenomeModelForTokenClassification = OmniModelForTokenClassification
OmniGenomeModelForSequenceRegression = OmniModelForSequenceRegression
OmniGenomeModelForTokenRegression = OmniModelForTokenRegression
OmniGenomeModelForStructuralImputation = OmniModelForStructuralImputation
OmniGenomeModelForMatrixRegression = OmniModelForMatrixRegression
OmniGenomeModelForMatrixClassification = OmniModelForMatrixClassification
OmniGenomeModelForMLM = OmniModelForMLM
OmniGenomeModelForSeq2Seq = OmniModelForSeq2Seq
OmniGenomeModelForRNADesign = OmniModelForRNADesign
OmniGenomeModelForEmbedding = OmniModelForEmbedding
OmniGenomeModelForAugmentation = OmniModelForAugmentation

# ------------------------------------------------------------------------------

__all__ = [
    "load_benchmark_datasets",
    "OmniDataset",
    "OmniModel",
    "OmniMetric",
    "OmniTokenizer",
    "OmniKmersTokenizer",
    "OmniSingleNucleotideTokenizer",
    "OmniBPETokenizer",
    "ModelHub",
    "Pipeline",
    "PipelineHub",
    "BenchHub",
    "AutoBench",
    "AutoTrain",
    "AutoConfig",
    "ClassificationMetric",
    "RegressionMetric",
    "RankingMetric",
    "Trainer",
    "HFTrainer",
    "AccelerateTrainer",
    "AutoBenchConfig",
    "download_benchmark",
    "download_model",
    "download_pipeline",
    "VoteEnsemblePredictor",
    "clean_temp_dir_pt_files",
    "fprint",
    "seed_everything",
    "save_args",
    "naive_secondary_structure_repair",
    "check_bench_version",
    "clean_temp_checkpoint",
    "print_args",
    "env_meta_info",
    "RNA2StructureCache",
    "OmniDatasetForSequenceClassification",
    "OmniDatasetForSequenceRegression",
    "OmniDatasetForTokenClassification",
    "OmniDatasetForTokenRegression",
    "OmniTokenizer",
    "OmniKmersTokenizer",
    "OmniSingleNucleotideTokenizer",
    "OmniBPETokenizer",
    "OmniDataset",
    "OmniMetric",
    "OmniModel",
    "OmniLoraModel",
    "OmniModelForSequenceClassification",
    "OmniModelForMultiLabelSequenceClassification",
    "OmniModelForTokenClassification",
    "OmniModelForSequenceRegression",
    "OmniModelForTokenRegression",
    "OmniModelForStructuralImputation",
    "OmniModelForMatrixRegression",
    "OmniModelForMatrixClassification",
    "OmniModelForMLM",
    "OmniModelForSeq2Seq",
    "OmniModelForRNADesign",
    "OmniModelForEmbedding",
    "OmniModelForAugmentation",
    # OmniGenome* aliases for backward compatibility
    "OmniGenomeTokenizer",
    "OmniGenomeKmersTokenizer",
    "OmniGenomeSingleNucleotideTokenizer",
    "OmniGenomeBPETokenizer",
    "OmniGenomeDataset",
    "OmniGenomeMetric",
    "OmniGenomeModel",
    "OmniGenomeDatasetForSequenceClassification",
    "OmniGenomeDatasetForSequenceRegression",
    "OmniGenomeDatasetForTokenClassification",
    "OmniGenomeDatasetForTokenRegression",
    "OmniGenomeLoraModel",
    "OmniGenomeModelForSequenceClassification",
    "OmniGenomeModelForMultiLabelSequenceClassification",
    "OmniGenomeModelForTokenClassification",
    "OmniGenomeModelForSequenceRegression",
    "OmniGenomeModelForTokenRegression",
    "OmniGenomeModelForStructuralImputation",
    "OmniGenomeModelForMatrixRegression",
    "OmniGenomeModelForMatrixClassification",
    "OmniGenomeModelForMLM",
    "OmniGenomeModelForSeq2Seq",
    "OmniGenomeModelForRNADesign",
    "OmniGenomeModelForEmbedding",
    "OmniGenomeModelForAugmentation",
    # Command line functions
    "run_bench",
    "bench_command",
    "run_train",
    "train_command",
    "EpistasisExplainer",
    "SequenceLogoExplainer",
    "Visualization2DExplainer",
]


LOGO1 = r"""
    **@@ #========= @@**            ___                     _
      **@@ +----- @@**             / _ \  _ __ ___   _ __  (_)
        **@@ = @@**               | | | || '_ ` _ \ | '_ \ | |
           **@@                   | |_| || | | | | || | | || |
        @@** = **@@                \___/ |_| |_| |_||_| |_||_|
     @@** ------+ **@@
   @@** =========# **@@            ____
  @@ ---------------+ @@          / ___|  ___  _ __
 @@ ================== @@        | |  _  / _ \| '_ \
  @@ +--------------- @@         | |_| ||  __/| | | |
   @@** #========= **@@           \____| \___||_| |_|
    @@** +------ **@@
       @@** = **@@
          @@**                    ____                      _
       **@@ = @@**               | __ )   ___  _ __    ___ | |__
    **@@ -----+  @@**            |  _ \  / _ \| '_ \  / __|| '_ \
  **@@ ==========# @@**          | |_) ||  __/| | | || (__ | | | |
  @@ --------------+ @@**        |____/  \___||_| |_| \___||_| |_|
"""

LOGO2 = r"""

   **  +----------- **           ___                     _
  @@                 @@         / _ \  _ __ ___   _ __  (_)
 @@* #============== *@@       | | | || '_ ` _ \ | '_ \ | |
 @@*                 *@@       | |_| || | | | | || | | || |
 *@@  +------------ *@@         \___/ |_| |_| |_||_| |_||_|
  *@*               @@*
   *@@  #========= @@*
    *@@*         *@@*
      *@@  +---@@@*              ____
        *@@*   **               / ___|  ___  _ __
          **@**                | |  _  / _ \| '_ \
        *@@* *@@*              | |_| ||  __/| | | |
      *@@ ---+  @@*             \____| \___||_| |_|
    *@@*         *@@*
   *@@ =========#  @@*
  *@@               @@*
 *@@ -------------+  @@*        ____                      _
 @@                   @@       | __ )   ___  _ __    ___ | |__
 @@ ===============#  @@       |  _ \  / _ \| '_ \  / __|| '_ \
  @@                 @@        | |_) ||  __/| | | || (__ | | | |
   ** -----------+  **         |____/  \___||_| |_| \___||_| |_|
"""

art_dna_color_map = {
    "*": "blue",  # Bases represented by '*'
    "@": "white",  # Bases represented by '@'
    "-": "yellow",  # Hydrogen bonds, assuming '-' represents a bond
    "=": "light_cyan",  # Hydrogen bonds, assuming '=' represents a bond
    "+": "yellow",  # '+' symbols in cyan
    " ": "black",  # Use red for undefined characters
}
import random

LOGO = random.choice([LOGO1, LOGO2])
print(LOGO)

clean_temp_dir_pt_files()



# -*- coding: utf-8 -*-
# file: utils.py
# time: 14:45 06/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.
import multiprocessing
import os
import pickle
import sys
import tempfile
import time
import warnings

import ViennaRNA as RNA
import findfile

default_omnigenome_repo = (
    "https://huggingface.co/spaces/yangheng/OmniGenomeLeaderboard/"
)


def seed_everything(seed=42):
    """
    Sets random seeds for reproducibility across all random number generators.

    This function sets seeds for Python's random module, NumPy, PyTorch (CPU and CUDA),
    and sets the PYTHONHASHSEED environment variable to ensure reproducible results
    across different runs.

    Args:
        seed (int): The seed value to use for all random number generators.
                   Defaults to 42.

    Example:
        >>> # Set seeds for reproducibility
        >>> seed_everything(42)
        >>> # Now all random operations will be reproducible
    """
    import random
    import numpy as np
    import torch

    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


class RNA2StructureCache(dict):
    """
    A cache for RNA secondary structure predictions using ViennaRNA.

    This class provides a caching mechanism for RNA secondary structure predictions
    to avoid redundant computations. It supports both single sequence and batch
    processing with optional multiprocessing for improved performance.

    Attributes:
        cache (dict): Dictionary storing sequence-structure mappings
        cache_file (str): Path to the cache file on disk
        queue_num (int): Counter for tracking cache updates
    """

    def __init__(self, cache_file=None, *args, **kwargs):
        """
        Initialize the RNA structure cache.

        Args:
            cache_file (str, optional): Path to the cache file. If None, uses
                                      a default temporary file.
            *args: Additional positional arguments for dict initialization
            **kwargs: Additional keyword arguments for dict initialization
        """
        super().__init__(*args, **kwargs)
        self.cache = dict(*args, **kwargs)
        self.cache_file = (
            cache_file
            if cache_file is not None
            else os.path.join(tempfile.gettempdir(), "rna_structure_cache.pkl")
        )
        self.queue_num = 0

        # Load existing cache if available
        if os.path.exists(self.cache_file):
            try:
                with open(self.cache_file, "rb") as f:
                    self.cache.update(pickle.load(f))
            except Exception as e:
                warnings.warn(f"Failed to load cache file: {e}")

    def __getitem__(self, key):
        """Gets a cached structure prediction."""
        return self.cache[key]

    def __setitem__(self, key, value):
        """Sets a structure prediction in the cache."""
        self.cache[key] = value

    def __str__(self):
        """String representation of the cache."""
        return str(self.cache)

    def __repr__(self):
        """String representation of the cache."""
        return str(self.cache)

    def _fold_single_sequence(self, sequence):
        """
        Predict structure for a single sequence (worker function for multiprocessing).

        Args:
            sequence (str): RNA sequence to fold

        Returns:
            tuple: (structure, mfe) tuple
        """
        try:
            return RNA.fold(sequence)
        except Exception as e:
            warnings.warn(f"Failed to fold sequence {sequence}: {e}")
            return ("." * len(sequence), 0.0)

    def fold(self, sequence, return_mfe=False, num_workers=1):
        """
        Predicts RNA secondary structure for given sequences.

        This method predicts RNA secondary structures using ViennaRNA. It supports
        both single sequences and batches of sequences. The method uses caching
        to avoid redundant predictions and supports multiprocessing for batch
        processing on non-Windows systems.

        Args:
            sequence (str or list): A single RNA sequence or a list of sequences.
            return_mfe (bool): Whether to return minimum free energy along with
                              structure. Defaults to False.
            num_workers (int): Number of worker processes for batch processing.
                             Defaults to 1. Set to None for auto-detection.

        Returns:
            str or list: The predicted structure(s). If return_mfe is True,
                        returns tuples of (structure, mfe).

        Example:
            >>> cache = RNA2StructureCache()
            >>> # Predict structure for a single sequence
            >>> structure = cache.fold("GGGAAAUCC")
            >>> print(structure)  # "(((...)))"

            >>> # Predict structures for multiple sequences
            >>> structures = cache.fold(["GGGAAAUCC", "AUUGCUAA"])
            >>> print(structures)  # ["(((...)))", "........"]
        """
        if not isinstance(sequence, list):
            sequences = [sequence]
        else:
            sequences = sequence

        # Determine if we should use multiprocessing
        use_multiprocessing = (
            os.name != "nt"  # Not Windows
            and len(sequences) > 1  # Multiple sequences
            and num_workers > 1  # Multiple workers requested
        )

        # Find sequences that need prediction
        sequences_to_predict = [seq for seq in sequences if seq not in self.cache]

        if sequences_to_predict:
            if use_multiprocessing:
                # Use multiprocessing for batch prediction
                if num_workers is None:
                    num_workers = min(os.cpu_count(), len(sequences_to_predict))

                try:
                    # Set multiprocessing start method to 'spawn' for better compatibility
                    if multiprocessing.get_start_method(allow_none=True) != "spawn":
                        multiprocessing.set_start_method("spawn", force=True)

                    with multiprocessing.Pool(num_workers) as pool:
                        # Use map instead of apply_async for better error handling
                        results = pool.map(
                            self._fold_single_sequence, sequences_to_predict
                        )

                        # Update cache with results
                        for seq, result in zip(sequences_to_predict, results):
                            self.cache[seq] = result
                            self.queue_num += 1

                except Exception as e:
                    warnings.warn(
                        f"Multiprocessing failed, falling back to sequential: {e}"
                    )
                    # Fallback to sequential processing
                    for seq in sequences_to_predict:
                        self.cache[seq] = self._fold_single_sequence(seq)
                        self.queue_num += 1
            else:
                # Sequential processing
                for seq in sequences_to_predict:
                    self.cache[seq] = self._fold_single_sequence(seq)
                    self.queue_num += 1

        # Prepare output
        if return_mfe:
            structures = [self.cache[seq] for seq in sequences]
        else:
            structures = [self.cache[seq][0] for seq in sequences]

        # Update cache file periodically
        self.update_cache_file(self.cache_file)

        # Return single result or list
        if len(structures) == 1:
            return structures[0]
        else:
            return structures

    def update_cache_file(self, cache_file=None):
        """
        Updates the cache file on disk.

        This method saves the in-memory cache to disk. It only saves when
        the queue_num reaches 100 to avoid excessive disk I/O.

        Args:
            cache_file (str, optional): Path to the cache file. If None, uses
                                      the instance's cache_file.

        Example:
            >>> cache.update_cache_file()  # Force save to disk
        """
        if self.queue_num < 100:
            return

        if cache_file is None:
            cache_file = self.cache_file

        try:
            if not os.path.exists(os.path.dirname(cache_file)):
                os.makedirs(os.path.dirname(cache_file))

            with open(cache_file, "wb") as f:
                pickle.dump(self.cache, f)

            self.queue_num = 0
        except Exception as e:
            warnings.warn(f"Failed to update cache file: {e}")


def env_meta_info():
    """
    Collects metadata about the current environment and library versions.

    This function gathers information about the current Python environment,
    including versions of key libraries like PyTorch and Transformers,
    as well as OmniGenBench version information.

    Returns:
        dict: A dictionary containing environment metadata including:
              - library_name: Name of the OmniGenBench library
              - omnigenbench_version: Version of OmniGenBench
              - torch_version: PyTorch version with CUDA info
              - transformers_version: Transformers library version

    Example:
        >>> metadata = env_meta_info()
        >>> print(metadata['torch_version'])  # "2.0.0+cu118+git..."
    """
    from torch.version import __version__ as torch_version
    from torch.version import cuda as torch_cuda_version
    from torch.version import git_version
    from transformers import __version__ as transformers_version
    from ... import __version__ as omnigenbench_version
    from ... import __name__ as omnigenbench_name

    return {
        "library_name": omnigenbench_name,
        "omnigenbench_version": omnigenbench_version,
        "torch_version": f"{torch_version}+cu{torch_cuda_version}+git{git_version}",
        "transformers_version": transformers_version,
    }


def naive_secondary_structure_repair(sequence, structure):
    """
    Repair the secondary structure of a sequence.

    This function attempts to repair malformed RNA secondary structure
    representations by ensuring proper bracket matching. It handles
    common issues like unmatched brackets by converting them to dots.

    Args:
        sequence (str): A string representing the sequence.
        structure (str): A string representing the secondary structure.

    Returns:
        str: A string representing the repaired secondary structure.

    Example:
        >>> sequence = "GGGAAAUCC"
        >>> structure = "(((...)"  # Malformed structure
        >>> repaired = naive_secondary_structure_repair(sequence, structure)
        >>> print(repaired)  # "(((...))"
    """
    repaired_structure = ""
    stack = []
    for i, (s, c) in enumerate(zip(structure, sequence)):
        if s == "(":
            stack.append(i)
        elif s == ")":
            if stack:
                stack.pop()
            else:
                repaired_structure += "."
        else:
            repaired_structure += s
    for i in stack:
        repaired_structure = repaired_structure[:i] + "." + repaired_structure[i + 1 :]
    return repaired_structure


def save_args(config, save_path):
    """
    Save arguments to a file.

    This function saves the arguments from a configuration object to a text file.
    It's useful for logging experiment parameters and configurations.

    Args:
        config: A Namespace object containing the arguments.
        save_path (str): A string representing the path of the file to be saved.

    Example:
        >>> from argparse import Namespace
        >>> config = Namespace(learning_rate=0.001, batch_size=32)
        >>> save_args(config, "config.txt")
    """
    f = open(os.path.join(save_path), mode="w", encoding="utf8")
    for arg in config.args:
        if config.args_call_count[arg]:
            f.write("{}: {}\n".format(arg, config.args[arg]))
    f.close()


def print_args(config, logger=None):
    """
    Print the arguments to the console.

    This function prints the arguments from a configuration object to the console
    or a logger. It's useful for debugging and logging experiment parameters.

    Args:
        config: A Namespace object containing the arguments.
        logger: A logger object. If None, prints to console.

    Example:
        >>> from argparse import Namespace
        >>> config = Namespace(learning_rate=0.001, batch_size=32)
        >>> print_args(config)
    """
    if logger is None:
        for arg in config.args:
            if config.args_call_count[arg]:
                print("{}: {}".format(arg, config.args[arg]))
    else:
        for arg in config.args:
            if config.args_call_count[arg]:
                logger.info("{}: {}".format(arg, config.args[arg]))


def fprint(*objects, sep=" ", end="\n", file=sys.stdout, flush=False):
    """
    Custom print function that adds a timestamp and the package version before the printed message.

    Args:
        *objects: Any number of objects to be printed
        sep (str, optional): Separator between objects. Defaults to " ".
        end (str, optional): Ending character after all objects are printed. Defaults to "\n".
        file (io.TextIOWrapper, optional): Text file to write printed output to. Defaults to sys.stdout.
        flush (bool, optional): Whether to flush output buffer after printing. Defaults to False.
    """
    from omnigenbench import __version__
    from omnigenbench import __name__

    print(
        time.strftime(
            "[%Y-%m-%d %H:%M:%S] [{} {}] ".format(__name__, __version__),
            time.localtime(time.time()),
        ),
        *objects,
        sep=sep,
        end=end,
        file=file,
        flush=flush,
    )


def clean_temp_checkpoint(days_threshold=7):
    """
    Clean up temporary checkpoint files older than specified days.

    This function removes temporary checkpoint files that are older than
    the specified threshold to free up disk space.

    Args:
        days_threshold (int): Number of days after which files are considered old.
                            Defaults to 7.

    Example:
        >>> clean_temp_checkpoint(3)  # Remove files older than 3 days
    """
    import glob
    import time

    temp_patterns = [
        "temp_checkpoint_*",
        "checkpoint_*",
        "*.tmp",
        "*.temp",
    ]

    current_time = time.time()
    threshold_time = current_time - (days_threshold * 24 * 60 * 60)

    for pattern in temp_patterns:
        for file_path in glob.glob(pattern):
            try:
                if os.path.getmtime(file_path) < threshold_time:
                    os.remove(file_path)
            except Exception:
                pass


def load_module_from_path(module_name, file_path):
    """
    Load a Python module from a file path.

    This function dynamically loads a Python module from a file path,
    useful for loading configuration files or custom modules.

    Args:
        module_name (str): Name to assign to the loaded module
        file_path (str): Path to the Python file to load

    Returns:
        module: The loaded module object

    Example:
        >>> config = load_module_from_path("config", "config.py")
        >>> print(config.some_variable)
    """
    import importlib.util

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def check_bench_version(bench_version, omnigenbench_version):
    """
    Check if benchmark version is compatible with OmniGenBench version.

    This function compares the benchmark version with the OmniGenBench version
    to ensure compatibility and warns if there are potential issues.

    Args:
        bench_version (str): Version of the benchmark
        omnigenbench_version (str): Version of OmniGenBench

    Example:
        >>> check_bench_version("0.2.0", "0.3.0")
    """

    if bench_version != omnigenbench_version:
        warnings.warn(
            f"Benchmark version ({bench_version}) differs from "
            f"OmniGenBench version ({omnigenbench_version}). "
            f"This may cause compatibility issues."
        )


def clean_temp_dir_pt_files():
    """
    Clean up temporary PyTorch files in the current directory.

    This function removes temporary PyTorch files (like .pt, .pth files)
    that may be left over from previous runs.

    Example:
        >>> clean_temp_dir_pt_files()
    """
    import glob

    temp_patterns = ["*.pt", "*.pth", "temp_*", "checkpoint_*"]

    for pattern in temp_patterns:
        for file_path in glob.glob(pattern):
            try:
                if os.path.isfile(file_path):
                    os.remove(file_path)
            except Exception:
                pass



# -*- coding: utf-8 -*-
# file: accelerate_trainer.py
# time: 14:40 06/04/2024
# author: YANG, HENG <hy345@exeter.ac.uk> (杨恒)
# github: https://github.com/yangheng95
# huggingface: https://huggingface.co/yangheng
# google scholar: https://scholar.google.com/citations?user=NPq5a_0AAAAJ&hl=en
# Copyright (C) 2019-2024. All Rights Reserved.
"""
Accelerate-based distributed training utilities.

This module provides HuggingFace Accelerate-based distributed training framework
for genomic models, including automatic mixed precision training, distributed
training support, early stopping, and model checkpointing.
"""

import os
import time
import numpy as np
from torch.utils.data import DataLoader
from tqdm import tqdm
from typing import Dict, Any, Optional, Union

import torch

from .base_trainer import BaseTrainer
from ..misc.utils import env_meta_info, fprint, seed_everything


class AccelerateTrainer(BaseTrainer):
    """
    HuggingFace Accelerate-based distributed trainer for genomic models.

    This trainer provides distributed training capabilities with automatic mixed precision,
    gradient accumulation, and early stopping. It supports both single and multi-GPU
    training with seamless integration with HuggingFace Accelerate.

    Attributes:
        accelerator: HuggingFace Accelerate instance for distributed training
        early_stop_flag: Tensor for coordinating early stopping across processes

    Example:
        >>> trainer = AccelerateTrainer(
        ...     model=model,
        ...     train_dataset=train_dataset,
        ...     eval_dataset=eval_dataset,
        ...     epochs=10,
        ...     batch_size=32,
        ...     optimizer=optimizer
        ... )
        >>> metrics = trainer.train()
    """

    def __init__(
        self,
        model: torch.nn.Module,
        **kwargs,
    ):
        """
        Initialize the accelerate trainer.

        Args:
            model (torch.nn.Module): The model to be trained
            **kwargs: Additional keyword arguments passed to BaseTrainer
        """
        super().__init__(model, **kwargs)

    def _setup_training_components(self) -> None:
        """
        Set up accelerate training-specific components.

        This method initializes the HuggingFace Accelerator with appropriate
        mixed precision settings and prepares the model, optimizer, and
        data loaders for distributed training.
        """
        # Determine mixed precision from `autocast` argument
        if self.autocast in ["float16", "fp16"]:
            mp_setting = "fp16"
        elif self.autocast in ["bfloat16", "bf16"]:
            mp_setting = "bf16"
        else:
            mp_setting = "no"

        # Initialize Accelerator
        from accelerate import Accelerator
        from accelerate import DistributedDataParallelKwargs

        ddp_kwargs = DistributedDataParallelKwargs(find_unused_parameters=True)
        self.accelerator = Accelerator(
            mixed_precision=mp_setting, kwargs_handlers=[ddp_kwargs]
        )

        # Prepare components with accelerator
        to_prepare = [self.model]
        if self.optimizer is not None:
            to_prepare.append(self.optimizer)
        if self.train_loader is not None:
            to_prepare.append(self.train_loader)
        if self.eval_loader is not None:
            to_prepare.append(self.eval_loader)
        if self.test_loader is not None:
            to_prepare.append(self.test_loader)

        prepared = self.accelerator.prepare(*to_prepare)
        self.model = prepared[0]

        idx = 1
        if self.optimizer is not None:
            self.optimizer = prepared[idx]
            idx += 1
        if self.train_loader is not None:
            self.train_loader = prepared[idx]
            idx += 1
        if self.eval_loader is not None:
            self.eval_loader = prepared[idx]
            idx += 1
        if self.test_loader is not None:
            self.test_loader = prepared[idx]

    def _prepare_batch(self, batch: Any) -> Any:
        """
        Prepare a batch for model input.

        For accelerate trainer, the batch is already prepared by accelerator,
        so we just return it as-is.

        Args:
            batch: Input batch

        Returns:
            The input batch (already prepared by accelerator)
        """
        return batch

    def _predict_batch(self, batch: Any) -> Dict[str, torch.Tensor]:
        """
        Generate predictions for a batch using the model.

        Args:
            batch: Input batch

        Returns:
            Dict[str, torch.Tensor]: Dictionary containing predictions
        """
        return self.accelerator.unwrap_model(self.model).predict(batch)

    def _train_epoch(self, epoch: int) -> float:
        """
        Train the model for one epoch using HuggingFace Accelerate.

        Args:
            epoch (int): Current epoch number

        Returns:
            float: Average training loss for the epoch
        """
        self.model.train()
        train_loss = []

        train_it = tqdm(
            self.train_loader,
            desc=f"Epoch {epoch + 1}/{self.epochs} Loss",
            disable=not self.accelerator.is_main_process,
        )

        # Training loop with accelerator.accumulate for gradient accumulation
        for step, batch in enumerate(train_it):
            with self.accelerator.accumulate(self.model):
                outputs = self.model(**batch)
                loss = self._compute_loss(outputs)

                # Backward pass and optimizer step within accumulate context
                self.accelerator.backward(loss)
                self.optimizer.step()
                self.optimizer.zero_grad()

            train_loss.append(loss.item())
            train_it.set_description(
                f"Epoch {epoch + 1}/{self.epochs} Loss: {np.nanmean(train_loss):.4f}"
            )

        return np.nanmean(train_loss)

    def _compute_loss(self, outputs: Dict[str, torch.Tensor]) -> torch.Tensor:
        """
        Compute loss from model outputs.

        Args:
            outputs (Dict[str, torch.Tensor]): Model outputs

        Returns:
            torch.Tensor: Computed loss

        Raises:
            ValueError: If no loss function is available
        """
        if "loss" in outputs:
            return outputs["loss"]

        # Try to use model's loss function
        if hasattr(self.model, "loss_function") and callable(self.model.loss_function):
            return self.model.loss_function(outputs["logits"], outputs["labels"])

        if (
            hasattr(self.model, "model")
            and hasattr(self.model.model, "loss_function")
            and callable(self.model.model.loss_function)
        ):
            return self.model.model.loss_function(outputs["logits"], outputs["labels"])

        raise ValueError(
            "The model does not have a loss function defined. "
            "Please provide a loss function or ensure the model has one."
        )

    def train(self, path_to_save: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Train the model using distributed training.

        This method performs the complete training loop with validation,
        early stopping, and model checkpointing. It handles distributed
        training across multiple GPUs and processes.

        Args:
            path_to_save (Optional[str]): Path to save the trained model
            **kwargs: Additional keyword arguments for model saving

        Returns:
            Dict[str, Any]: Dictionary containing training metrics
        """
        seed_everything(self.seed)

        # Initialize early stopping flag for distributed coordination
        self.early_stop_flag = torch.tensor(0, device=self.accelerator.device)

        # Ensure all processes sync before starting
        self.accelerator.wait_for_everyone()

        # Initial evaluation
        if self.eval_loader is not None and len(self.eval_loader) > 0:
            initial_metrics = self.evaluate()
        else:
            initial_metrics = self.test()

        # Only main process handles metric comparison and model saving
        if self.accelerator.is_main_process:
            if self._is_metric_better(initial_metrics, stage="valid"):
                self._save_state_dict()
                self.early_stop_flag = torch.tensor(0, device=self.accelerator.device)

        # Synchronize early stopping flag across all processes
        gathered_flags = self.accelerator.gather(self.early_stop_flag)
        self.early_stop_flag = (
            gathered_flags if gathered_flags.ndim == 0 else gathered_flags[0]
        )

        # Main training loop
        for epoch in range(self.epochs):
            # Train for one epoch
            avg_loss = self._train_epoch(epoch)

            # Synchronize all processes before evaluation
            self.accelerator.wait_for_everyone()

            # Evaluate after each epoch
            if self.eval_loader is not None and len(self.eval_loader) > 0:
                valid_metrics = self.evaluate()
            else:
                valid_metrics = self.test()

            # Only main process handles metric comparison and early stopping
            if self.accelerator.is_main_process:
                if self._is_metric_better(valid_metrics, stage="valid"):
                    self._save_state_dict()
                    self.early_stop_flag = torch.tensor(
                        0, device=self.accelerator.device
                    )
                else:
                    self.early_stop_flag += 1

            # Synchronize early stopping flag across all processes
            gathered_flags = self.accelerator.gather(self.early_stop_flag)
            self.early_stop_flag = (
                gathered_flags if gathered_flags.ndim == 0 else gathered_flags[0]
            )

            # Check for early stopping
            if self.early_stop_flag.item() > self.patience:
                if self.accelerator.is_main_process:
                    fprint(f"Early stopping at epoch {epoch + 1}.")
                break

            # Save epoch checkpoint (only main process)
            if path_to_save and self.accelerator.is_main_process:
                self._save_epoch_checkpoint(
                    path_to_save, epoch, valid_metrics, **kwargs
                )

            # Ensure all processes sync before next epoch
            self.accelerator.wait_for_everyone()

        # Final testing with best model
        if self.test_loader is not None and len(self.test_loader) > 0:
            self._load_state_dict()
            self.accelerator.wait_for_everyone()
            test_metrics = self.test()
            if self.accelerator.is_main_process:
                self._is_metric_better(test_metrics, stage="test")

        # Save final model (only main process)
        if path_to_save and self.accelerator.is_main_process:
            self._save_final_model(path_to_save, **kwargs)

        # Clean up
        self._remove_state_dict()
        ## DO NOT REMOVE THE FOLLOWING BLOCK, OTHERWISE AUTOBENCH WILL NOT WORK ##
        self.accelerator.free_memory(
            self.model,
            self.optimizer,
            self.train_loader,
            self.eval_loader,
            self.test_loader,
        )
        ## Remove accelerator reference to avoid memory leaks ##
        # delattr(self, "accelerator")
        return self.metrics

    def evaluate(self) -> Dict[str, Any]:
        """
        Evaluate the model on the validation dataset.

        This method runs the model in evaluation mode and computes metrics
        on the validation dataset. It handles distributed evaluation and
        gathers results from all processes.

        Returns:
            Dict[str, Any]: Dictionary containing evaluation metrics
        """
        self.model.eval()
        all_truth = []
        all_preds = []

        # Disable progress bar on non-main processes
        it = tqdm(
            self.eval_loader,
            desc="Evaluating",
            disable=not self.accelerator.is_main_process,
        )

        with torch.no_grad():
            for batch in it:
                output = self.accelerator.unwrap_model(self.model).predict(batch)
                predictions = output["predictions"]
                labels = batch["labels"]

                # Gather predictions and labels from all processes
                gathered_predictions = self.accelerator.gather(predictions)
                gathered_labels = self.accelerator.gather(labels)

                # Only main process processes gathered data
                if self.accelerator.is_main_process:
                    gathered_predictions = gathered_predictions.float().cpu().numpy()
                    gathered_labels = gathered_labels.float().cpu().numpy()
                    all_preds.append(gathered_predictions)
                    all_truth.append(gathered_labels)

        # Only main process computes metrics
        if self.accelerator.is_main_process:
            all_preds = np.concatenate(all_preds, axis=0)
            all_truth = np.concatenate(all_truth, axis=0)

            if not np.all(all_truth == -100):
                valid_metrics = {}
                for metric_func in self.compute_metrics:
                    valid_metrics.update(metric_func(all_truth, all_preds))
            else:
                valid_metrics = {
                    "Validation labels predictions may be NaN. No metrics calculated.": 0
                }

            fprint(valid_metrics)
        else:
            valid_metrics = None

        self.predictions.update({"valid": {"pred": all_preds, "true": all_truth}})
        return valid_metrics

    def test(self) -> Dict[str, Any]:
        """
        Test the model on the test dataset.

        This method runs the model in evaluation mode and computes metrics
        on the test dataset. It handles distributed testing and gathers
        results from all processes.

        Returns:
            Dict[str, Any]: Dictionary containing test metrics
        """
        self.model.eval()
        all_truth = []
        all_preds = []

        it = tqdm(
            self.test_loader,
            desc="Testing",
            disable=not self.accelerator.is_main_process,
        )

        with torch.no_grad():
            for batch in it:
                output = self.accelerator.unwrap_model(self.model).predict(batch)
                predictions = output["predictions"]
                labels = batch["labels"]

                gathered_predictions = self.accelerator.gather(predictions)
                gathered_labels = self.accelerator.gather(labels)

                if self.accelerator.is_main_process:
                    gathered_predictions = gathered_predictions.float().cpu().numpy()
                    gathered_labels = gathered_labels.float().cpu().numpy()
                    all_preds.append(gathered_predictions)
                    all_truth.append(gathered_labels)

        # Only main process computes metrics
        if self.accelerator.is_main_process:
            all_preds = np.concatenate(all_preds, axis=0)
            all_truth = np.concatenate(all_truth, axis=0)

            if not np.all(all_truth == -100):
                test_metrics = {}
                for metric_func in self.compute_metrics:
                    test_metrics.update(metric_func(all_truth, all_preds))
            else:
                test_metrics = {
                    "Test labels predictions may be NaN. No metrics calculated.": 0
                }
            fprint(test_metrics)
        else:
            test_metrics = None

        self.predictions.update({"test": {"pred": all_preds, "true": all_truth}})
        return test_metrics

    def _is_metric_better(self, metrics: Dict[str, Any], stage: str = "valid") -> bool:
        """
        Check if the current metrics are better than the best metrics so far.

        This method is overridden to ensure only the main process performs
        metric comparison in distributed training.

        Args:
            metrics (Dict[str, Any]): Current metrics
            stage (str): Stage of evaluation ('valid' or 'test')

        Returns:
            bool: True if current metrics are better, False otherwise
        """
        # Only main process performs metric comparison
        if not self.accelerator.is_main_process:
            return False

        return super()._is_metric_better(metrics, stage)

    def save_model(self, path: str, overwrite: bool = False, **kwargs) -> None:
        """
        Save the trained model.

        Args:
            path (str): Path to save the model
            overwrite (bool): Whether to overwrite existing files (default: False)
            **kwargs: Additional keyword arguments for model saving
        """
        # Only main process saves the model
        if not hasattr(self, "accelerator"):
            self.model.save(path, overwrite, **kwargs)
        elif self.accelerator.is_main_process:
            self.accelerator.unwrap_model(self.model).save(path, overwrite, **kwargs)

    def _load_state_dict(self) -> None:
        """Load the best model state dictionary."""
        if hasattr(self, "_model_state_dict_path") and os.path.exists(
            self._model_state_dict_path
        ):
            weights = torch.load(self._model_state_dict_path, map_location="cpu")
            self.accelerator.unwrap_model(self.model).load_state_dict(weights)

    def _save_state_dict(self) -> None:
        """Save the current model state dictionary."""
        if not hasattr(self, "_model_state_dict_path"):
            from hashlib import sha256

            time_str = time.strftime("%Y%m%d_%H%M%S", time.localtime())
            hash_digest = sha256(self.__repr__().encode("utf-8")).hexdigest()
            self._model_state_dict_path = f"tmp_ckpt_{time_str}_{hash_digest}.pt"

        if os.path.exists(self._model_state_dict_path):
            os.remove(self._model_state_dict_path)

        # Only main process saves the state dict
        if self.accelerator.is_main_process:
            torch.save(
                self.accelerator.unwrap_model(self.model).state_dict(),
                self._model_state_dict_path,
            )

    def _remove_state_dict(self) -> None:
        """Remove the temporary model state dictionary file."""
        if not hasattr(self, "_model_state_dict_path"):
            from hashlib import sha256

            time_str = time.strftime("%Y%m%d_%H%M%S", time.localtime())
            hash_digest = sha256(self.__repr__().encode("utf-8")).hexdigest()
            self._model_state_dict_path = f"tmp_ckpt_{time_str}_{hash_digest}.pt"

        if (
            os.path.exists(self._model_state_dict_path)
            and self.accelerator.is_main_process
        ):
            os.remove(self._model_state_dict_path)
